# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""Phius Certification Data Class"""

from copy import copy

try:
    from typing import Optional, Union
except ImportError:
    pass  # IronPython 2.7

from honeybee_ph import _base
from honeybee_ph_utils import enumerables

# -----------------------------------------------------------------------------


class PhiusBuildingCertificationProgram(enumerables.CustomEnum):
    allowed = [
        "1-Default",
        "2-PHIUS 2015",
        "3-PHIUS 2018",
        "4-Italian",
        "5-PHIUS 2018 CORE",
        "6-PHIUS 2018 ZERO",
        "7-PHIUS 2021 CORE",
        "8-PHIUS 2021 ZERO",
    ]

    def __init__(self, _value=1):
        # type: (Union[str, int]) -> None
        super(PhiusBuildingCertificationProgram, self).__init__(_value)


class PhiusBuildingCategoryType(enumerables.CustomEnum):
    allowed = [
        "1-RESIDENTIAL BUILDING",
        "2-NON-RESIDENTIAL BUILDING",
    ]

    def __init__(self, _value=1):
        # type: (Union[str, int]) -> None
        super(PhiusBuildingCategoryType, self).__init__(_value)


class PhiusBuildingUseType(enumerables.CustomEnum):
    allowed = [
        "1-RESIDENTIAL",
        "",
        "",
        "4-OFFICE/ADMINISTRATIVE BUILDING",
        "5-SCHOOL",
        "6-OTHER",
        "7-UNDEFINED/UNFINISHED",
    ]

    def __init__(self, _value=1):
        # type: (Union[str, int]) -> None
        super(PhiusBuildingUseType, self).__init__(_value)

        # Quick double check cus' the numbering here is so stupid
        if self.value == "":
            raise Exception("Error: WUFI's '_building_use_type' numbering is weird. Check the 'inputs' value is valid?")


class PhiusBuildingStatus(enumerables.CustomEnum):
    allowed = [
        "1-IN_PLANNING",
        "2-UNDER_CONSTRUCTION",
        "3-COMPLETE",
    ]

    def __init__(self, _value=1):
        # type: (Union[str, int]) -> None
        super(PhiusBuildingStatus, self).__init__(_value)


class PhiusBuildingType(enumerables.CustomEnum):
    allowed = [
        "1-NEW_CONSTRUCTION",
        "2-RETROFIT",
        "3-MIXED",
    ]

    def __init__(self, _value=1):
        # type: (Union[str, int]) -> None
        super(PhiusBuildingType, self).__init__(_value)


# -----------------------------------------------------------------------------


class PhiusCertification(_base._Base):
    def __init__(self):
        super(PhiusCertification, self).__init__()
        self.localization_selection_type = 2

        self._certification_program = PhiusBuildingCertificationProgram("7-PHIUS 2021 CORE")
        self._building_category_type = PhiusBuildingCategoryType("1-RESIDENTIAL BUILDING")
        self._building_use_type = PhiusBuildingUseType("1-RESIDENTIAL")
        self._building_status = PhiusBuildingStatus("1-IN_PLANNING")
        self._building_type = PhiusBuildingType("1-NEW_CONSTRUCTION")

        self.PHIUS2021_heating_demand = 15.0
        self.PHIUS2021_cooling_demand = 15.0
        self.PHIUS2021_heating_load = 10.0
        self.PHIUS2021_cooling_load = 10.0

        self.int_gains_evap_per_person = 15
        self.int_gains_flush_heat_loss = True
        self.int_gains_num_toilets = 1
        self.int_gains_toilet_room_util_pat = None
        self.int_gains_use_school_defaults = False
        self.int_gains_dhw_marginal_perf_ratio = None

        self.icfa_override = None  # type: Optional[float]

    @property
    def certification_program(self):
        # type: () -> PhiusBuildingCertificationProgram
        return self._certification_program

    @certification_program.setter
    def certification_program(self, _input):
        # type: (Optional[Union[str, int]]) -> None
        if _input:
            self._certification_program = PhiusBuildingCertificationProgram(_input)

    @property
    def building_category_type(self):
        # type: () -> PhiusBuildingCategoryType
        return self._building_category_type

    @building_category_type.setter
    def building_category_type(self, _input):
        # type: (Optional[Union[str, int]]) -> None
        if _input:
            self._building_category_type = PhiusBuildingCategoryType(_input)

    @property
    def building_use_type(self):
        # type: () -> PhiusBuildingUseType
        return self._building_use_type

    @building_use_type.setter
    def building_use_type(self, _input):
        # type: (Optional[Union[str, int]]) -> None
        if _input:
            self._building_use_type = PhiusBuildingUseType(_input)

    @property
    def building_status(self):
        # type: () -> PhiusBuildingStatus
        return self._building_status

    @building_status.setter
    def building_status(self, _input):
        # type: (Optional[Union[str, int]]) -> None
        if _input:
            self._building_status = PhiusBuildingStatus(_input)

    @property
    def building_type(self):
        # type: () -> PhiusBuildingType
        return self._building_type

    @building_type.setter
    def building_type(self, _input):
        # type: (Optional[Union[str, int]]) -> None
        if _input:
            self._building_type = PhiusBuildingType(_input)

    def __str__(self):
        return "{}()".format(self.__class__.__name__)

    def __repr__(self):
        return str(self)

    def ToString(self):
        return str(self)

    def to_dict(self):
        # type: () -> dict
        d = {}

        d["localization_selection_type"] = self.localization_selection_type

        d["certification_program"] = self.certification_program.to_dict()
        d["building_category_type"] = self.building_category_type.to_dict()
        d["building_use_type"] = self.building_use_type.to_dict()
        d["building_status"] = self.building_status.to_dict()
        d["building_type"] = self.building_type.to_dict()

        d["PHIUS2021_heating_demand"] = self.PHIUS2021_heating_demand
        d["PHIUS2021_cooling_demand"] = self.PHIUS2021_cooling_demand
        d["PHIUS2021_heating_load"] = self.PHIUS2021_heating_load
        d["PHIUS2021_cooling_load"] = self.PHIUS2021_cooling_load

        d["int_gains_evap_per_person"] = self.int_gains_evap_per_person
        d["int_gains_flush_heat_loss"] = self.int_gains_flush_heat_loss
        d["int_gains_num_toilets"] = self.int_gains_num_toilets
        d["int_gains_toilet_room_util_pat"] = self.int_gains_toilet_room_util_pat
        d["int_gains_use_school_defaults"] = self.int_gains_use_school_defaults
        d["int_gains_dhw_marginal_perf_ratio"] = self.int_gains_dhw_marginal_perf_ratio

        d["identifier"] = self.identifier
        d["user_data"] = copy(self.user_data)
        d["display_name"] = self.display_name

        d["icfa_override"] = self.icfa_override

        return d

    @classmethod
    def from_dict(cls, _dict):
        # type: (dict) -> PhiusCertification
        obj = cls()

        obj.localization_selection_type = _dict.get("localization_selection_type")

        obj._certification_program = PhiusBuildingCertificationProgram.from_dict(_dict.get("certification_program", {}))
        obj._building_category_type = PhiusBuildingCategoryType.from_dict(_dict.get("building_category_type", {}))
        obj._building_use_type = PhiusBuildingUseType.from_dict(_dict.get("building_use_type", {}))
        obj._building_status = PhiusBuildingStatus.from_dict(_dict.get("building_status", {}))
        obj._building_type = PhiusBuildingType.from_dict(_dict.get("building_type", {}))

        obj.PHIUS2021_heating_demand = _dict.get("PHIUS2021_heating_demand")
        obj.PHIUS2021_cooling_demand = _dict.get("PHIUS2021_cooling_demand")
        obj.PHIUS2021_heating_load = _dict.get("PHIUS2021_heating_load")
        obj.PHIUS2021_cooling_load = _dict.get("PHIUS2021_cooling_load")

        obj.int_gains_evap_per_person = _dict.get("int_gains_evap_per_person")
        obj.int_gains_flush_heat_loss = _dict.get("int_gains_flush_heat_loss")
        obj.int_gains_num_toilets = _dict.get("int_gains_num_toilets")
        obj.int_gains_toilet_room_util_pat = _dict.get("int_gains_toilet_room_util_pat")
        obj.int_gains_use_school_defaults = _dict.get("int_gains_use_school_defaults")
        obj.int_gains_dhw_marginal_perf_ratio = _dict.get("int_gains_dhw_marginal_perf_ratio")

        obj.identifier = _dict.get("identifier")
        obj.user_data = _dict.get("user_data", {})
        obj.display_name = _dict.get("display_name")

        obj.icfa_override = _dict.get("icfa_override")

        return obj

    def __copy__(self):
        # type: () -> PhiusCertification
        obj = PhiusCertification()

        obj.set_base_attrs_from_source(self)
        obj.localization_selection_type = self.localization_selection_type

        obj._certification_program = self._certification_program
        obj._building_category_type = self._building_category_type
        obj._building_use_type = self._building_use_type
        obj._building_status = self._building_status
        obj._building_type = self._building_type

        obj.PHIUS2021_heating_demand = self.PHIUS2021_heating_demand
        obj.PHIUS2021_cooling_demand = self.PHIUS2021_cooling_demand
        obj.PHIUS2021_heating_load = self.PHIUS2021_heating_load
        obj.PHIUS2021_cooling_load = self.PHIUS2021_cooling_load

        obj.int_gains_evap_per_person = self.int_gains_evap_per_person
        obj.int_gains_flush_heat_loss = self.int_gains_flush_heat_loss
        obj.int_gains_num_toilets = self.int_gains_num_toilets
        obj.int_gains_toilet_room_util_pat = self.int_gains_toilet_room_util_pat
        obj.int_gains_use_school_defaults = self.int_gains_use_school_defaults
        obj.int_gains_dhw_marginal_perf_ratio = self.int_gains_dhw_marginal_perf_ratio

        obj.icfa_override = self.icfa_override

        return obj

    def duplicate(self):
        # type: () -> PhiusCertification
        return self.__copy__()



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""Calculate Window U-w as per ISO-10077-2:2006"""

import warnings

try:
    from ladybug_geometry.geometry3d import LineSegment3D, face
except ImportError as e:
    raise ImportError("Failed to import ladybug_geometry from ladybug_geometry.")

try:
    from honeybee import aperture
except ImportError as e:
    raise ImportError("Failed to import honeybee from honeybee.")

try:
    from honeybee_energy_ph.construction import window
except ImportError as e:
    raise ImportError("Failed to import honeybee_energy_ph from honeybee_energy_ph.")


def get_ladybug_Face3D_width_and_height(_lbt_face3d):
    # type: (face.Face3D) -> tuple[float, float]
    """Get the width and height of a ladybug_geometry.geometry3d.face.Face3D object.

    If a non-rectangular object with more than 4 vertices is passed, a ValueError is raised.
    """

    if len(_lbt_face3d.vertices) != 4:
        raise ValueError(
            "The LBT-Face3D '{}' has {} vertices. Only use rectangular faces with 4 vertices.".format(
                _lbt_face3d, len(_lbt_face3d.vertices)
            )
        )

    (
        vertix_upper_left,
        vertix_upper_right,
        vertix_lower_right,
        vertix_lower_left,
    ) = _lbt_face3d.upper_left_counter_clockwise_vertices
    horizontal_edge = LineSegment3D.from_end_points(vertix_upper_left, vertix_upper_right)
    vertical_edge = LineSegment3D.from_end_points(vertix_upper_right, vertix_lower_right)

    height = horizontal_edge.length
    width = vertical_edge.length

    return width, height


def get_honeybee_aperture_width_and_height(_hb_aperture):
    # type: (aperture.Aperture) -> tuple[float, float]
    """Get the width and height of a rectangular honeybee.aperture.Aperture object.

    If a non-rectangular object with more than 4 vertices is passed, a ValueError is raised.
    """
    if len(_hb_aperture.geometry.vertices) != 4:
        raise ValueError(
            "The HB-Aperture '{}' has {} vertices. Only use rectangular Apertures with 4 vertices.".format(
                _hb_aperture.display_name, len(_hb_aperture.geometry.vertices)
            )
        )

    return get_ladybug_Face3D_width_and_height(_hb_aperture.geometry)


class ISO100771Data:
    """Wrapper class used to calculate the ISO-10077-1 heat-loss values."""

    def __init__(self, _win_width, _win_height, _frame, _glazing):
        # type: (float, float, window.PhWindowFrame, window.PhWindowGlazing) -> None
        self.win_width = _win_width
        self.win_height = _win_height
        self.frame = _frame
        self.glazing = _glazing

    @classmethod
    def from_hb_aperture(cls, _hb_aperture):
        # type: (aperture.Aperture) -> ISO100771Data
        """Create an ISO100771Data object from a honeybee.aperture.Aperture object."""

        hbe_prop = _hb_aperture.properties.energy  # type: ignore

        try:
            ph_frame = hbe_prop.construction.properties.ph.ph_frame
        except Exception as e:
            raise Exception("The Aperture does not have a PH-Style frame.", e)

        try:
            ph_glazing = hbe_prop.construction.properties.ph.ph_glazing
        except Exception as e:
            raise Exception("The Aperture does not have a PH-Style glazing.", e)

        (
            w,
            h,
        ) = get_honeybee_aperture_width_and_height(_hb_aperture)
        return cls(
            w,
            h,
            ph_frame,
            ph_glazing,
        )

    @classmethod
    def from_lbt_Face3D(cls, _hb_face3d, _ph_frame, _ph_glazing):
        # type: (face.Face3D, window.PhWindowFrame, window.PhWindowGlazing) -> ISO100771Data
        """Create an ISO100771Data object from a ladybug_geometry.geometry3d.face.Face3D object."""

        (
            w,
            h,
        ) = get_ladybug_Face3D_width_and_height(_hb_face3d)
        return cls(
            w,
            h,
            _ph_frame,
            _ph_glazing,
        )

    @property
    def area_window(self):
        # type: () -> float
        return self.side_exterior_length("right") * self.side_exterior_length("top")

    @property
    def area_glazing(self):
        # type: () -> float
        """Return the area of the glazing."""
        return self.side_interior_length("right") * self.side_interior_length("top")

    @property
    def area_frame(self):
        # type: () -> float
        """Return the total area of all the frames."""
        return self.area_window - self.area_glazing

    def frame_element(self, _side):
        # type: (str) -> window.PhWindowFrameElement
        """Return a specific frame element for the given side."""
        return getattr(self.frame, _side)

    def side_interior_length(self, _side):
        # type: (str) -> float
        side_lengths = {
            "top": (self.win_width - self.frame.right.width - self.frame.left.width),
            "right": (self.win_height - self.frame.top.width - self.frame.bottom.width),
            "bottom": (self.win_width - self.frame.right.width - self.frame.left.width),
            "left": (self.win_height - self.frame.top.width - self.frame.bottom.width),
        }
        return side_lengths[_side]

    def side_exterior_length(self, _side):
        # type: (str) -> float
        side_lengths = {
            "top": self.win_width,
            "right": self.win_height,
            "bottom": self.win_width,
            "left": self.win_height,
        }
        return side_lengths[_side]

    def get_adjacent_frame(self, _side):
        # type: (str) -> tuple[window.PhWindowFrameElement, window.PhWindowFrameElement]
        """Get the adjacent frames for the given side (ie: top -> (left, right) )."""
        adjacent_frames = {
            "top": {"adj_1": "left", "adj_2": "right"},
            "right": {"adj_1": "top", "adj_2": "bottom"},
            "bottom": {"adj_1": "left", "adj_2": "right"},
            "left": {"adj_1": "top", "adj_2": "bottom"},
        }
        adjacent_frame_1 = getattr(self.frame, adjacent_frames[_side]["adj_1"])  # type: window.PhWindowFrameElement
        adjacent_frame_2 = getattr(self.frame, adjacent_frames[_side]["adj_2"])  # type: window.PhWindowFrameElement
        return adjacent_frame_1, adjacent_frame_2

    def corner_area(self, _side):
        # type: (str) -> float
        """Get the area at the corners of the frames.

        This will return 1/2 to the total area of both corners.
        """
        base_frame = self.frame_element(_side)
        adjacent_frame_1, adjacent_frame_2 = self.get_adjacent_frame(_side)
        corner_area_1 = base_frame.width * adjacent_frame_1.width
        corner_area_2 = base_frame.width * adjacent_frame_2.width
        corner_area = ((corner_area_1) / 2) + ((corner_area_2) / 2)
        return corner_area

    def side_area(self, _side):
        # type: (str) -> float
        center_area = self.frame_element(_side).width * self.side_interior_length(_side)
        return center_area + self.corner_area(_side)

    def side_frame_heat_loss(self, _side):
        # type: (str) -> float
        return self.frame_element(_side).u_factor * self.side_area(_side)

    def side_psi_glazing_heat_lost(self, _side):
        # type: (str) -> float
        return self.frame_element(_side).psi_glazing * self.side_interior_length(_side)

    def side_psi_install_heat_lost(self, _side):
        # type: (str) -> float
        return self.frame_element(_side).psi_install * self.side_exterior_length(_side)

    @property
    def heat_loss_frame(self):
        # type: () -> float
        return sum(
            [
                self.side_frame_heat_loss("top"),
                self.side_frame_heat_loss("right"),
                self.side_frame_heat_loss("bottom"),
                self.side_frame_heat_loss("left"),
            ]
        )

    @property
    def heat_loss_psi_glazing(self):
        # type: () -> float
        return sum(
            [
                self.side_psi_glazing_heat_lost("top"),
                self.side_psi_glazing_heat_lost("right"),
                self.side_psi_glazing_heat_lost("bottom"),
                self.side_psi_glazing_heat_lost("left"),
            ]
        )

    @property
    def heat_loss_psi_install(self):
        # type: () -> float
        return sum(
            [
                self.side_psi_install_heat_lost("top"),
                self.side_psi_install_heat_lost("right"),
                self.side_psi_install_heat_lost("bottom"),
                self.side_psi_install_heat_lost("left"),
            ]
        )

    @property
    def heat_loss_glazing(self):
        # type: () -> float
        return self.area_glazing * self.glazing.u_factor

    @property
    def uw(self):
        # type: () -> float
        return (
            self.heat_loss_glazing + self.heat_loss_frame + self.heat_loss_psi_glazing + self.heat_loss_psi_install
        ) / self.area_window

    @property
    def wufi_passive_uw(self):
        # type: () -> float
        """Implemented following the WUFI-Passive C# method as closely as possible.

        Note that in this case, the top and bottom frame areas are extended to the corners,
        while the left and right frame areas are cut short at the corners. This is different than
        the `uw` method implemented above which cuts the corners at a 45° angle.
        """

        # --
        glazing_edge_length = 2 * (
            self.win_width
            - self.frame.left.width
            - self.frame.right.width
            + self.win_height
            - self.frame.bottom.width
            - self.frame.top.width
        )

        # ---
        glazing_edge_left = self.win_height - self.frame.bottom.width - self.frame.top.width
        glazing_edge_right = self.win_height - self.frame.bottom.width - self.frame.top.width
        glazing_edge_top = self.win_width - self.frame.left.width - self.frame.right.width
        glazing_edge_bottom = self.win_width - self.frame.left.width - self.frame.right.width
        average_psi_glazing_w_m_k = (
            (self.frame.left.psi_glazing * glazing_edge_left)
            + (self.frame.right.psi_glazing * glazing_edge_right)
            + (self.frame.top.psi_glazing * glazing_edge_top)
            + (self.frame.bottom.psi_glazing * glazing_edge_bottom)
        ) / (glazing_edge_left + glazing_edge_right + glazing_edge_top + glazing_edge_bottom)

        # ---
        install_edge_length = (self.win_height * 2) + (self.win_width * 2)
        average_psi_frame_w_m_k = (
            (self.frame.left.psi_install * self.win_height)
            + (self.frame.right.psi_install * self.win_height)
            + (self.frame.top.psi_install * self.win_width)
            + (self.frame.bottom.psi_install * self.win_width)
        ) / install_edge_length

        # ---
        frame_area_left = glazing_edge_left * self.frame.left.width
        frame_area_right = glazing_edge_right * self.frame.right.width
        frame_area_top = self.win_width * self.frame.top.width
        frame_area_bottom = self.win_width * self.frame.bottom.width
        total_frame_area = frame_area_left + frame_area_right + frame_area_top + frame_area_bottom
        average_u_value_frame_w_m2k = (
            (self.frame.left.u_factor * frame_area_left)
            + (self.frame.right.u_factor * frame_area_right)
            + (self.frame.top.u_factor * frame_area_top)
            + (self.frame.bottom.u_factor * frame_area_bottom)
        ) / total_frame_area

        # ---
        window_area = self.win_width * self.win_height
        glazing_area = glazing_edge_left * glazing_edge_top
        frame_area = window_area - glazing_area
        return (
            (self.glazing.u_factor * glazing_area)
            + (average_u_value_frame_w_m2k * frame_area)
            + (average_psi_glazing_w_m_k * glazing_edge_length)
            + (average_psi_frame_w_m_k * install_edge_length)
        ) / window_area


def build_standard_window(_frame, _glazing):
    # type: (window.PhWindowFrame, window.PhWindowGlazing) -> ISO100771Data
    """A standard-size (1.23m x 1.48m) window as per ISO 10077-2:2006, As per Annex F (2006)"""
    width = 1.23  # M
    height = 1.48  # M
    return ISO100771Data(width, height, _frame, _glazing)


def calculate_window_frame_factor(_frame, _glazing):
    # type: (window.PhWindowFrame, window.PhWindowGlazing) -> float
    """Calculate the frame factor for a standard-size (1.23m x 1.48m) window as per ISO 10077-2:2006, Annex-F"""
    window = build_standard_window(_frame, _glazing)
    return window.area_glazing / window.area_window


# -----------------------------------------------------------------------------
# -- Functions to calculate U-w -----------------------------------------------


def calculate_window_uw(_frame, _glazing):
    # type: (window.PhWindowFrame, window.PhWindowGlazing) -> float
    """Calculate U-w (W/m2k) for a standard-size (1.23m x 1.48m) window as per ISO 10077-2:2006, Annex-F
    This value includes the impact of the frame, glazing, glazing-spacers, and the psi-install heat loss.
    """

    warnings.warn(
        "The 'function calculate_window_uw' is deprecated and will be removed in a future version. Use function 'calculate_standard_window_uw' instead.",
        DeprecationWarning,
    )

    window = build_standard_window(_frame, _glazing)
    return window.uw


def calculate_standard_window_uw(_frame, _glazing):
    # type: (window.PhWindowFrame, window.PhWindowGlazing) -> float
    """Calculate U-w (W/m2k) for a standard-size window as per ISO 10077-2:2006
    This value includes the impact of the frame, glazing, glazing-spacers, and the psi-install heat loss.
    """

    window = build_standard_window(_frame, _glazing)
    return window.uw


def calculate_hb_aperture_uw(_hb_aperture):
    # type: (aperture.Aperture) -> float
    """Calculate U-W (W/m2k) for a 'honeybee.aperture.Aperture' with PH-Style frame and glazing as per ISO 10077-2:2006
    This value includes the impact of the frame, glazing, glazing-spacers, and the psi-install heat loss.
    """
    window = ISO100771Data.from_hb_aperture(_hb_aperture)
    return window.uw


def calculate_lbt_Face3D_uw(_hb_face3d, _ph_frame, _ph_glazing):
    # type: (face.Face3D, window.PhWindowFrame, window.PhWindowGlazing) -> float
    """Calculate U-w (W/m2k) for a 'ladybug_geometry.geometry3d.face.Face3D' as per ISO 10077-2:2006

    This value includes the impact of the frame, glazing, glazing-spacers, and the psi-install heat loss.
    Note that in this case, the frame-areas are calculated by splitting them at a 45°
    angle at the corner. This is slightly different than the WUFI-Passive implementation that
    extends the top and bottom frames, while cutting the left and right frames short.
    """

    window = ISO100771Data.from_lbt_Face3D(_hb_face3d, _ph_frame, _ph_glazing)
    return window.uw



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""Tools for working with Ladybug.geometry3d.Vector3D objects."""

import math

try:
    from typing import Union
except ImportError:
    pass  # Python 2.7

try:
    from ladybug_geometry.geometry3d.pointvector import Vector2D, Vector3D
except ImportError as e:
    raise ImportError("Failed to import ladybug_geometry: " + str(e))


def vector_equal(vector_a, vector_b, _tolerance):
    # type: (Vector3D, Vector3D, float) -> bool
    """Return True if two Vector3Ds are equal within tolerance."""

    return all(abs(vector_a[i] - vector_b[i]) <= _tolerance for i in range(len(vector_a)))


def cross_product(vector_a, vector_b):
    # type: (Vector3D, Vector3D) -> Vector3D
    """Return the cross product of two Vector3Ds."""

    return Vector3D(
        vector_a.y * vector_b.z - vector_a.z * vector_b.y,  # type: ignore
        vector_a.z * vector_b.x - vector_a.x * vector_b.z,  # type: ignore
        vector_a.x * vector_b.y - vector_a.y * vector_b.x,  # type: ignore
    )


def dot_product(vector_a, vector_b):
    # type: (Union[Vector2D, Vector3D], Union[Vector2D, Vector3D]) -> float
    """Return the dot product of two Vector3Ds."""

    return sum(vector_a[i] * vector_b[i] for i in range(len(vector_a)))


def magnitude(vector):
    # type: (Vector3D) -> float
    """Return the magnitude of a Vector3D."""

    return sum(vector[i] ** 2 for i in range(len(vector))) ** 0.5


def normalize(vector):
    # type: (Vector3D) -> Vector3D
    """Return a normalized Vector3D."""

    mag = magnitude(vector)
    try:
        return Vector3D(*[vector[i] / mag for i in range(len(vector))])
    except ZeroDivisionError:
        return Vector3D(0, 0, 0)


def angle_between_2D_vectors(_vector1, _vector2):
    # type: (Union[Vector3D, Vector2D], Union[Vector3D, Vector2D]) -> float
    """Return the angle (in radians) between two 2D vectors."""

    # Calculate the dot product between the X-axis vectors
    cos_theta = dot_product(_vector1, _vector2)

    # Ensure cos_theta is within [-1, 1] to avoid invalid arc-cos input
    cos_theta = max(min(cos_theta, 1), -1)

    # Calculate the angle in radians
    theta_radians = math.acos(cos_theta)

    return theta_radians



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""HB-PH-HVAC Model Properties."""

try:
    from typing import Any, Dict, Optional
except ImportError:
    pass  # Python 2.7

try:
    from itertools import izip as zip  # type: ignore
except ImportError:
    pass  # Python3

try:
    from honeybee import extensionutil, model
except ImportError as e:
    raise ImportError("\nFailed to import honeybee:\n\t{}".format(e))

try:
    from honeybee_phhvac.heat_pumps import PhHeatPumpSystemBuilder
    from honeybee_phhvac.heating import PhHeatingSystemBuilder
    from honeybee_phhvac.hot_water_system import PhHotWaterSystem
    from honeybee_phhvac.renewable_devices import PhRenewableEnergyDeviceBuilder
    from honeybee_phhvac.supportive_device import PhSupportiveDevice
    from honeybee_phhvac.ventilation import PhExhaustDeviceBuilder, PhVentilationSystem
except ImportError as e:
    raise ImportError("\nFailed to import honeybee_phhvac:\n\t{}".format(e))


class ModelPhHvacProperties(object):
    def __init__(self, _host):
        # type: (Optional[model.Model]) -> None
        self._host = _host

    @property
    def host(self):
        # type: () -> Optional[model.Model]
        return self._host

    def duplicate(self, new_host=None):
        # type: (Any) -> ModelPhHvacProperties
        _host = new_host or self._host
        new_properties_obj = ModelPhHvacProperties(_host)
        return new_properties_obj

    def to_dict(self, abridged=False):
        # type: (bool) -> dict[str, dict]
        d = {}
        if abridged == False:
            d["type"] = "ModelPhHvacProperties"
        else:
            d["type"] = "ModelPhHvacPropertiesAbridged"
        return {"ph_hvac": d}

    @classmethod
    def from_dict(cls, _dict, host):
        # type: (dict[str, Any], Any) -> ModelPhHvacProperties
        if _dict["type"] not in ("ModelPhHvacProperties", "ModelPhHvacPropertiesAbridged"):
            raise ValueError(
                "Expected ModelPhHvacProperties or ModelPhHvacPropertiesAbridged. Got {}.".format(_dict["type"])
            )
        new_prop = cls(host)
        return new_prop

    @staticmethod
    def _build_mechanical_devices_from_dict(data):
        # type: (list[dict]) -> dict[str, dict[str, Any]]
        """Return a dict of the mechanical systems, keyed by system-type then identifier."""

        mechanical_systems = {
            "ventilation_systems": {},
            "heating_systems": {},
            "heat_pump_systems": {},
            "exhaust_vent_devices": {},
            "supportive_devices": {},
            "renewable_devices": {},
            "hot_water_systems": {},
        }

        for room_dict in data:
            d = room_dict.get("ventilation_system", {})
            if d:
                mechanical_systems["ventilation_systems"][d["identifier"]] = PhVentilationSystem.from_dict(d)

            for d in room_dict.get("heating_systems", []):
                mechanical_systems["heating_systems"][d["identifier"]] = PhHeatingSystemBuilder.from_dict(d)

            for d in room_dict.get("heat_pump_systems", []):
                mechanical_systems["heat_pump_systems"][d["identifier"]] = PhHeatPumpSystemBuilder.from_dict(d)

            for d in room_dict.get("supportive_devices", []):
                mechanical_systems["supportive_devices"][d["identifier"]] = PhSupportiveDevice.from_dict(d)

            for d in room_dict.get("exhaust_vent_devices", []):
                mechanical_systems["exhaust_vent_devices"][d["identifier"]] = PhExhaustDeviceBuilder.from_dict(d)

            for d in room_dict.get("renewable_devices", []):
                mechanical_systems["renewable_devices"][d["identifier"]] = PhRenewableEnergyDeviceBuilder.from_dict(d)

            d = room_dict.get("hot_water_system", {})
            if d:
                mechanical_systems["hot_water_systems"][d["identifier"]] = PhHotWaterSystem.from_dict(d)

        return mechanical_systems

    @staticmethod
    def load_properties_from_dict(data):
        # type: (Dict[str, Dict]) -> None
        """Load the ModelPhHvacProperties attributes from an HB-Model dictionary as Python objects.

        Loaded objects include: None

        The function is called when re-serializing an HB-Model object from a
        dictionary. It will load honeybee_ph entities as Python objects and returns
        a tuple of dictionaries with all the de-serialized Honeybee-PH objects.

        Arguments:
        ----------
            data: A dictionary representation of an entire honeybee-core Model.
                 Note that this dictionary must have ModelPhProperties
                in order for this method to successfully apply the .ph properties.

                Note: data is an HB-Model dict and .keys() will include:
                [
                    'display_name', 'identifier', 'tolerance',
                    'angle_tolerance', 'rooms', 'type', 'version',
                    'units', 'orphaned_shades', 'properties'
                ]
        """
        assert "ph_hvac" in data["properties"], "Error: HB-Model Dictionary possesses no ModelPhHvacProperties?"

        return None

    def apply_properties_from_dict(self, data):
        # type: (Dict[str, Any]) -> None
        """Apply the ".ph_hvac" properties of a dictionary to the host Model of this object.

        This method is called when the HB-Model is de-serialized from a dict back into
        a Python object. In an 'Abridged' HBJSON file, all the property information
        is stored at the model level, not at the sub-model object level. In that case,
        this method is used to apply the right property data back onto all the sub-model
        objects (faces, rooms, apertures, etc).

        Arguments:
        ----------
            * data (dict[str, Any]): A dictionary representation of an entire honeybee-core
                Model. Note that this dictionary must have ModelPhHvacProperties
                in order for this method to successfully apply the .ph properties.

                Note: data is an HB-Model dict and .keys() will include:
                [
                    'display_name', 'identifier', 'tolerance',
                    'angle_tolerance', 'rooms', 'type', 'version',
                    'units', 'orphaned_shades', 'properties'
                ]

        Returns:
        --------
            * None
        """

        if not self.host:
            return None

        assert "ph_hvac" in data["properties"], "Error: Dictionary possesses no ModelPhHvacProperties?"

        # -------------------------------------------------------------------------------
        # -- 1) Re-build all of the 'ModelPhHvac' objects from the HB-Model dict as python objects
        _ = self.load_properties_from_dict(data)

        # -------------------------------------------------------------------------------
        # -- 2) Collect lists of .ph_hvac property dictionaries at the sub-model level (room, face, etc)
        (
            room_ph_dicts,
            face_ph_dicts,
            shd_ph_dicts,
            ap_ph_dicts,
            dr_ph_dicts,
        ) = extensionutil.model_extension_dicts(data, "ph_hvac", [], [], [], [], [])

        # -------------------------------------------------------------------------------
        # -- 3) Re-Build all of the mechanical HVAC objects from the HB-Model dict as python objects
        mechanical_systems = self._build_mechanical_devices_from_dict(room_ph_dicts)

        # -------------------------------------------------------------------------------
        # -- Pass the Mechanical Devices to the Rooms
        for room, room_dict in zip(self.host.rooms, room_ph_dicts):
            if not room_dict:
                continue
            room.properties.ph_hvac.apply_properties_from_dict(room_dict, mechanical_systems)

        # -- Pull out all the Apertures, Faces, Shades, and Doors from the HB-Model
        apertures, faces, shades, doors = [], [], [], []
        for hb_room in self.host.rooms:
            for face in hb_room.faces:
                faces.append(face)
                for aperture in face.apertures:
                    apertures.append(aperture)

        for face, face_dict in zip(faces, face_ph_dicts):
            face.properties.ph_hvac.apply_properties_from_dict(face_dict)

        for aperture, ap_dict in zip(apertures, ap_ph_dicts):
            aperture.properties.ph_hvac.apply_properties_from_dict(ap_dict)

        for shade, ap_dict in zip(shades, shd_ph_dicts):
            shade.properties.ph_hvac.apply_properties_from_dict(ap_dict)

        for door, ap_dict in zip(doors, dr_ph_dicts):
            door.properties.ph_hvac.apply_properties_from_dict(ap_dict)



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""Tools for working with Honeybee.face.Face objects."""


import math
from collections import defaultdict

try:
    from typing import List, Sequence, TypeVar, Union
except ImportError:
    pass  # IronPython 2.7

try:
    T = TypeVar("T", bound="Union[face.Face, shade.Shade]")
except Exception:
    pass  # IronPython 2.7

try:
    from honeybee import face, shade
except ImportError as e:
    raise ImportError("Failed to import honeybee")

try:
    from ladybug_geometry.geometry3d import pointvector
except ImportError:
    raise ImportError("Failed to import ladybug_geometry")


def _hb_face_type_unique_key(_hb_face):
    # type: (face.Face | shade.Shade) -> str
    """Return a unique key for an HB-Face's type considering the type, construction, and bc."""
    return "{}_{}_{}".format(
        getattr(_hb_face, "type", "shade"),
        getattr(_hb_face, "boundary_condition", "shade"),
        _hb_face.properties.energy.construction.display_name,  # type: ignore
    )


def sort_hb_faces_by_type(_faces):
    # type: (Sequence[T]) -> List[List[T]]
    """Group HB-Faces by their type."""

    d = defaultdict(list)
    for face in _faces:
        key = _hb_face_type_unique_key(face)
        d[key].append(face.duplicate())
    return list(d.values())


def sort_hb_faces_by_co_planar(_faces, _tolerance, _angle_tolerance_radians):
    # type: (List[T], float, float) -> List[List[T]]
    """Group HB-Faces with their co-planar neighbors.
    Args:
        _faces: (List[Face | Shade]) A list of HB-Faces to sort.
        _tolerance: (Model units) The tolerance value for co-planarity test, in model units.
        _angle_tolerance: (Radians) The tolerance for co-planarity, in radians.
    Returns:
        (List[List[Face | Shade]]) A list of lists of HB-Faces that are co-planar.
    """

    groups = {}
    for face in _faces:
        for group_plane, group_faces in groups.items():
            if face.geometry.plane.is_coplanar_tolerance(group_plane, _tolerance, _angle_tolerance_radians):
                group_faces.append(face)
                break
        else:
            groups[face.geometry.plane] = [face]
    return list(groups.values())


def points_are_coincident(_pt1, _pt2, _tolerance):
    # type: (pointvector.Point3D, pointvector.Point3D, float) -> bool
    """Return True if two Point3D objects are coincident within the tolerance."""
    return _pt1.distance_to_point(_pt2) < _tolerance


def hb_faces_are_touching(_face_2, _face_1, _tolerance):
    # type: (face.Face, face.Face, float) -> bool
    """Return True if the faces are 'touching' one another within the tolerance."""

    for v in _face_1.vertices:
        if _face_2.geometry.is_point_on_face(v, _tolerance):
            return True
        elif any([points_are_coincident(v, v2, _tolerance) for v2 in _face_2.vertices]):
            return True
    return False


def find_connected_HB_Faces(_hb_faces, _tolerance):
    # type: (List[face.Face], float) -> List[List[face.Face]]
    """Finds 'connected' HB-Faces.

    Args:
        _hb_faces List[Face | Shade]: A list of Honeybee face or shades to search for connected components.
        _tolerance: A tolerance value for determining whether two faces are touching.

    Returns:
        A list of lists, where each inner list contains a connected component of touching faces.
    """

    """Initialize an empty set called visited to keep track of which faces have 
    been visited during the search, and an empty list called components to store 
    the connected component groups."""
    visited = set()
    components = []

    def depth_first_search(node, component):
        # type: (face.Face, List[face.Face]) -> None
        """Define a recursive function that takes a starting face node
        and a list component to store the connected component.
        The function adds the starting face to the visited set and the component list,
        and then recursively calls itself on all neighboring faces that are not
        in the visited set and are touching the starting face within
        a given tolerance _tolerance.
        """
        visited.add(node)
        component.append(node)

        for _neighbor_face in _hb_faces:
            if _neighbor_face not in visited and hb_faces_are_touching(node, _neighbor_face, _tolerance):
                depth_first_search(_neighbor_face, component)

    """Loop over all the faces in the input list _hb_faces. If a face has not 
    been visited yet, create an empty list called 'component', call the 'depth_first_search'
    function with the face and the new component list, and append the 'component' list 
    to the master components list."""
    for hb_face in _hb_faces:
        if hb_face not in visited:
            component = []
            depth_first_search(hb_face, component)
            components.append(component)

    return components


def group_hb_faces(_hb_faces, _tolerance, _angle_tolerance_degrees):
    # type: (Sequence[T], float, float) -> List[List[T]]
    """Sort HB-Faces into groups of similar, planar, connected faces.

    Args:
        _hb_faces: (List[Face | Shade]) A list of HB-Faces to sort.
        _tolerance: (Model units) The tolerance value for co-planarity test, in model units.
        _angle_tolerance_degrees: (Degrees) The tolerance for co-planarity, in degrees.
    Returns:
        (List[List[Face | Shade]]) A list of lists of HB-Faces that are similar, planar, and connected.
    """

    face_groups_by_type = sort_hb_faces_by_type(_hb_faces)
    angle_tolerance_radians = math.radians(_angle_tolerance_degrees)

    face_groups_coplanar = []
    for face_group in face_groups_by_type:
        face_groups_coplanar.extend(sort_hb_faces_by_co_planar(face_group, _tolerance, angle_tolerance_radians))

    face_groups_connected = []
    for face_group in face_groups_coplanar:
        face_groups_connected.extend(find_connected_HB_Faces(face_group, _tolerance))

    return face_groups_connected



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""Process Equipment PH-Properties"""

try:
    from typing import Any
except:
    pass  # IronPython


try:
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from honeybee_energy.properties.extension import ProcessProperties
except ImportError as e:
    pass  # IronPython


try:
    from honeybee_energy_ph.load.ph_equipment import PhEquipment, PhEquipmentBuilder
except ImportError as e:
    raise ImportError('Failed to import honeybee_energy_ph', e)


class ProcessPhProperties_FromDictError(Exception):
    def __init__(self, _expected_types, _input_type):
        self.msg = 'Error: Expected type of "{}". Got: {}'.format(_expected_types, _input_type)
        super(ProcessPhProperties_FromDictError, self).__init__(self.msg)


class ProcessPhProperties(object):
    def __init__(self, _host):
        # type: (ProcessProperties) -> None
        self._host = _host
        self._ph_equipment = None

    @property
    def ph_equipment(self):
        # type: () -> PhEquipment | None
        return self._ph_equipment
    
    @ph_equipment.setter
    def ph_equipment(self, _equipment):
        # type: (PhEquipment) -> None
        if not isinstance(_equipment, PhEquipment):
            raise ValueError("Input must be of type PhEquipment")
        self._ph_equipment = _equipment

    @property
    def host(self):
        # type: () -> ProcessProperties
        return self._host

    def to_dict(self, abridged=False):
        # type: (bool) -> dict
        d = {}

        if abridged:
            d["type"] = "ProcessPhPropertiesAbridged"
        else:
            d["type"] = "ProcessPhProperties"
        
        if self._ph_equipment:
            d["equipment"] = self._ph_equipment.to_dict(_abridged=abridged)

        return {"ph": d}

    @classmethod
    def from_dict(cls, _input_dict, _host):
        # type: (dict, Any) -> ProcessPhProperties
        valid_types = (
            "ProcessPhProperties",
            "ProcessPhPropertiesAbridged",
        )
        if _input_dict["type"] not in valid_types:
            raise ProcessPhProperties_FromDictError(valid_types, _input_dict["type"])

        new_prop = cls(_host)

        if "equipment" in _input_dict:
            new_prop.ph_equipment = PhEquipmentBuilder.from_dict(_input_dict["equipment"], _host=new_prop)

        return new_prop

    def apply_properties_from_dict(self, abridged_data):
        # type: (dict) -> None
        return None

    def __copy__(self, new_host=None):
        # type: (ProcessProperties | None) -> ProcessPhProperties
        host = new_host or self._host
        new_obj = self.__class__(host)
        if self._ph_equipment:
            new_obj.ph_equipment = PhEquipmentBuilder.from_dict(self._ph_equipment.to_dict(), _host=new_obj)
        return new_obj

    def duplicate(self, new_host=None):
        # type: (Any) -> ProcessPhProperties
        return self.__copy__(new_host)

    def __str__(self):
        return "{}()".format(self.__class__.__name__)

    def __repr__(self):
        return str(self)

    def ToString(self):
        return str(self)



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""DEPRECATED in favor of the new 'honeybee_phhvac' module."""


from honeybee_energy_ph.hvac import _base
from honeybee_ph_utils import enumerables


class PhPipeDiameter(enumerables.CustomEnum):
    message = "PhPipeDiameter is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)


class PhPipeMaterial(enumerables.CustomEnum):
    message = "PhPipeMaterial is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)


class PhPipeSegment(_base._PhHVACBase):
    message = "PhPipeSegment is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)


class PhPipeElement(_base._PhHVACBase):
    message = "PhPipeElement is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)


class PhPipeBranch(_base._PhHVACBase):
    message = "PhPipeBranch is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)


class PhPipeTrunk(_base._PhHVACBase):
    message = "PhPipeTrunk is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)


class PhSHWTankType(enumerables.CustomEnum):
    message = "PhSHWTankType is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)


class PhSHWTank(_base._PhHVACBase):
    message = "PhSHWTank is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)


class PhHotWaterHeater(_base._PhHVACBase):
    message = "PhHotWaterHeater is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)


class PhSHWHeaterElectric(PhHotWaterHeater):
    message = "PhSHWHeaterElectric is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)


class PhSHWHeaterBoiler(PhHotWaterHeater):
    message = "PhSHWHeaterBoiler is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)


class PhSHWHeaterBoilerWood(PhHotWaterHeater):
    message = "PhSHWHeaterBoilerWood is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)


class PhSHWHeaterDistrict(PhHotWaterHeater):
    message = "PhSHWHeaterDistrict is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)


class PhSHWHeaterHeatPump(PhHotWaterHeater):
    message = "PhSHWHeaterHeatPump is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)


class PhSHWHeaterBuilder(object):
    message = "PhSHWHeaterBuilder is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""DEPRECATED in favor of the new 'honeybee_phhvac' module."""


class _PhHVACBase(object):
    message = "_PhHVACBase is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""Honeybee-PH-HVAC: Hot Water Devices."""

from copy import copy

try:
    from typing import Any, Dict, Union
except ImportError:
    pass  # IronPython

try:
    from honeybee_phhvac import _base
except ImportError as e:
    raise ImportError("Failed to import honeybee_phhvac", e)

try:
    from honeybee_ph_utils import enumerables
except ImportError as e:
    raise ImportError("Failed to import honeybee_ph_utils", e)


class UnknownPhHvacHotWaterHeaterTypeError(Exception):
    def __init__(self, _heater_types, _received_type):
        self.msg = 'Error: Unknown SHW Heater Type? Got: "{}" but only types: {} are allowed?'.format(
            _received_type, _heater_types
        )
        super(UnknownPhHvacHotWaterHeaterTypeError, self).__init__(self.msg)


# -- Tank ---------------------------------------------------------------------


class PhHvacHotWaterTankType(enumerables.CustomEnum):
    allowed = ["0-No storage tank", "1-DHW and heating", "2-DHW only"]

    def __init__(self, _value=1, _index_offset=0):
        # type: (Union[str, int], int) -> None
        super(PhHvacHotWaterTankType, self).__init__(_value, _index_offset)


class PhHvacHotWaterTank(_base._PhHVACBase):
    def __init__(self):
        super(PhHvacHotWaterTank, self).__init__()
        self.display_name = "_unnamed_hw_tank_"  # type: str
        self.quantity = 1  # type: int
        self._tank_type = PhHvacHotWaterTankType("2-DHW only")  # type: enumerables.CustomEnum
        self.in_conditioned_space = True  # type: bool
        self.solar_connection = False  # type: bool
        self.solar_losses = 0.0  # type: float
        self.storage_loss_rate = 0.0  # type: float
        self.storage_capacity = 300  # type: float
        self.standby_losses = 4.0  # type: float
        self.standby_fraction = 0.30  # type: float
        self.room_temp = 20.0  # type: float
        self.water_temp = 60.0  # type: float

    @property
    def tank_type(self):
        # type () -> str
        return self._tank_type.value

    @tank_type.setter
    def tank_type(self, _in):
        # type: (Union[str, int]) -> None
        self._tank_type = PhHvacHotWaterTankType(_in)

    def __copy__(self):
        # type: () -> PhHvacHotWaterTank
        new_obj = PhHvacHotWaterTank()

        new_obj.display_name = self.display_name
        new_obj.identifier = self.identifier
        new_obj.user_data = copy(self.user_data)
        new_obj.quantity = self.quantity
        new_obj._tank_type = self._tank_type
        new_obj.in_conditioned_space = self.in_conditioned_space
        new_obj.solar_connection = self.solar_connection
        new_obj.solar_losses = self.solar_losses
        new_obj.storage_loss_rate = self.storage_loss_rate
        new_obj.storage_capacity = self.storage_capacity
        new_obj.standby_losses = self.standby_losses
        new_obj.standby_fraction = self.standby_fraction
        new_obj.room_temp = self.room_temp
        new_obj.water_temp = self.water_temp

        return new_obj

    def duplicate(self):
        # type: () -> PhHvacHotWaterTank
        return self.__copy__()

    def to_dict(self, _include_properties=False):
        # type: (bool) -> dict[str, Any]
        d = super(PhHvacHotWaterTank, self).to_dict()
        d["quantity"] = self.quantity
        d["_tank_type"] = self._tank_type.to_dict()
        d["in_conditioned_space"] = self.in_conditioned_space
        d["solar_connection"] = self.solar_connection
        d["solar_losses"] = self.solar_losses
        d["storage_loss_rate"] = self.storage_loss_rate
        d["storage_capacity"] = self.storage_capacity
        d["standby_losses"] = self.standby_losses
        d["standby_fraction"] = self.standby_fraction
        d["room_temp"] = self.room_temp
        d["water_temp"] = self.water_temp

        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (dict[str, Any]) -> PhHvacHotWaterTank
        obj = cls()

        obj.display_name = _input_dict["display_name"]
        obj.identifier = _input_dict["identifier"]
        obj.user_data = _input_dict["user_data"]
        obj.quantity = _input_dict["quantity"]
        obj._tank_type = PhHvacHotWaterTankType.from_dict(_input_dict["_tank_type"])
        obj.in_conditioned_space = _input_dict["in_conditioned_space"]
        obj.solar_connection = _input_dict["solar_connection"]
        obj.solar_losses = _input_dict["solar_losses"]
        obj.storage_loss_rate = _input_dict["storage_loss_rate"]
        obj.storage_capacity = _input_dict["storage_capacity"]
        obj.standby_losses = _input_dict["standby_losses"]
        obj.standby_fraction = _input_dict["standby_fraction"]
        obj.room_temp = _input_dict["room_temp"]
        obj.water_temp = _input_dict["water_temp"]

        return obj

    def __str__(self):
        return "{}: {} {}".format(self.__class__.__name__, self.tank_type, self.display_name)

    def __repr__(self):
        return "{}: {}".format(self.__class__.__name__, self.display_name)

    def ToString(self):
        return self.__repr__()


# -- Heaters ------------------------------------------------------------------


class PhHvacHotWaterHeater(_base._PhHVACBase):
    """Base class for all PH Hot-Water Heaters."""

    def __init__(self):
        super(PhHvacHotWaterHeater, self).__init__()
        self.percent_coverage = 1.0
        self.in_conditioned_space = True

    def __copy__(self):
        # type: () -> PhHvacHotWaterHeater
        # TODO: Implement copy on all the subclasses
        return self

    def duplicate(self):
        # type: () -> PhHvacHotWaterHeater
        return self.__copy__()

    def to_dict(self, _include_properties=False):
        # type: (bool) -> dict
        d = super(PhHvacHotWaterHeater, self).to_dict()
        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict[str, Any]) -> PhHvacHotWaterHeater
        new_obj = cls()
        new_obj.identifier = _input_dict["identifier"]
        new_obj.display_name = _input_dict["display_name"]
        new_obj.user_data = _input_dict.get("user_data", {})
        new_obj.percent_coverage = _input_dict["percent_coverage"]
        new_obj.in_conditioned_space = _input_dict["in_conditioned_space"]
        return new_obj

    def __str__(self):
        return "{}()".format(self.__class__.__name__)

    def __repr__(self):
        return str(self)

    def ToString(self):
        return str(self)


class PhHvacHotWaterHeaterElectric(PhHvacHotWaterHeater):
    def __init__(self):
        super(PhHvacHotWaterHeaterElectric, self).__init__()

    def to_dict(self, _include_properties=False):
        # type: (bool) -> Dict[str, Any]
        d = super(PhHvacHotWaterHeaterElectric, self).to_dict()
        d["heater_type"] = self.__class__.__name__
        d["percent_coverage"] = self.percent_coverage
        d["in_conditioned_space"] = self.in_conditioned_space
        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict[str, Any]) -> PhHvacHotWaterHeaterElectric
        new_obj = cls()
        new_obj.identifier = _input_dict["identifier"]
        new_obj.display_name = _input_dict["display_name"]
        new_obj.user_data = _input_dict.get("user_data", {})
        new_obj.percent_coverage = _input_dict["percent_coverage"]
        new_obj.in_conditioned_space = _input_dict["in_conditioned_space"]
        return new_obj


class PhHvacHotWaterHeaterBoiler(PhHvacHotWaterHeater):
    def __init__(self):
        super(PhHvacHotWaterHeaterBoiler, self).__init__()
        self.fuel = 1  # Gas
        self.condensing = True
        self.effic_at_30_perc_load = 0.98
        self.effic_at_nominal_load = 0.94
        self.avg_return_temp_at_30_perc_load = 30
        self.avg_boiler_temp_at_70_55 = 41
        self.avg_boiler_temp_at_55_45 = 35
        self.avg_boiler_temp_at_35_28 = 24

    def to_dict(self, _include_properties=False):
        # type: (bool) -> dict
        d = super(PhHvacHotWaterHeaterBoiler, self).to_dict()
        d["heater_type"] = self.__class__.__name__
        d["percent_coverage"] = self.percent_coverage
        d["in_conditioned_space"] = self.in_conditioned_space
        d["fuel"] = self.fuel
        d["condensing"] = self.condensing
        d["effic_at_30_perc_load"] = self.effic_at_30_perc_load
        d["effic_at_nominal_load"] = self.effic_at_nominal_load
        d["avg_return_temp_at_30_perc_load"] = self.avg_return_temp_at_30_perc_load
        d["avg_boiler_temp_at_70_55"] = self.avg_boiler_temp_at_70_55
        d["avg_boiler_temp_at_55_45"] = self.avg_boiler_temp_at_55_45
        d["avg_boiler_temp_at_35_28"] = self.avg_boiler_temp_at_35_28
        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (dict) -> PhHvacHotWaterHeaterBoiler
        new_obj = cls()
        new_obj.identifier = _input_dict["identifier"]
        new_obj.display_name = _input_dict["display_name"]
        new_obj.user_data = _input_dict.get("user_data", {})
        new_obj.percent_coverage = _input_dict["percent_coverage"]
        new_obj.in_conditioned_space = _input_dict["in_conditioned_space"]
        new_obj.fuel = _input_dict["fuel"]
        new_obj.condensing = _input_dict["condensing"]
        new_obj.effic_at_30_perc_load = _input_dict["effic_at_30_perc_load"]
        new_obj.effic_at_nominal_load = _input_dict["effic_at_nominal_load"]
        new_obj.avg_return_temp_at_30_perc_load = _input_dict["avg_return_temp_at_30_perc_load"]
        new_obj.avg_boiler_temp_at_70_55 = _input_dict["avg_boiler_temp_at_70_55"]
        new_obj.avg_boiler_temp_at_55_45 = _input_dict["avg_boiler_temp_at_55_45"]
        new_obj.avg_boiler_temp_at_35_28 = _input_dict["avg_boiler_temp_at_35_28"]
        return new_obj


class PhHvacHotWaterHeaterBoilerWood(PhHvacHotWaterHeater):
    def __init__(self):
        # type: () -> None
        super(PhHvacHotWaterHeaterBoilerWood, self).__init__()
        self.fuel = 1  # Pellet
        self.effic_in_basic_cycle = 0.72
        self.effic_in_const_operation = 0.80
        self.avg_frac_heat_released = 0.5
        self.on_off_temp_diff = 30

    def to_dict(self, _include_properties=False):
        # type: (bool) -> Dict[str, Any]
        d = super(PhHvacHotWaterHeaterBoilerWood, self).to_dict()
        d["heater_type"] = self.__class__.__name__
        d["percent_coverage"] = self.percent_coverage
        d["in_conditioned_space"] = self.in_conditioned_space
        d["fuel"] = self.fuel
        d["effic_in_basic_cycle"] = self.effic_in_basic_cycle
        d["effic_in_const_operation"] = self.effic_in_const_operation
        d["avg_frac_heat_released"] = self.avg_frac_heat_released
        d["on_off_temp_diff"] = self.on_off_temp_diff
        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict[str, Any]) -> PhHvacHotWaterHeaterBoilerWood
        new_obj = cls()
        new_obj.identifier = _input_dict["identifier"]
        new_obj.display_name = _input_dict["display_name"]
        new_obj.user_data = _input_dict.get("user_data", {})
        new_obj.percent_coverage = _input_dict["percent_coverage"]
        new_obj.in_conditioned_space = _input_dict["in_conditioned_space"]
        new_obj.fuel = _input_dict["fuel"]
        new_obj.effic_in_basic_cycle = _input_dict["effic_in_basic_cycle"]
        new_obj.effic_in_const_operation = _input_dict["effic_in_const_operation"]
        new_obj.avg_frac_heat_released = _input_dict["avg_frac_heat_released"]
        new_obj.on_off_temp_diff = _input_dict["on_off_temp_diff"]
        return new_obj


class PhHvacHotWaterHeaterDistrict(PhHvacHotWaterHeater):
    def __init__(self):
        super(PhHvacHotWaterHeaterDistrict, self).__init__()
        self.energy_carrier = 1
        self.solar_fraction = 0
        self.util_fact_heat_transfer = 1

    def to_dict(self, _include_properties=False):
        # type: (bool) -> Dict[str, Any]
        d = super(PhHvacHotWaterHeaterDistrict, self).to_dict()
        d["heater_type"] = self.__class__.__name__
        d["percent_coverage"] = self.percent_coverage
        d["in_conditioned_space"] = self.in_conditioned_space
        d["energy_carrier"] = self.energy_carrier
        d["solar_fraction"] = self.solar_fraction
        d["util_fact_heat_transfer"] = self.util_fact_heat_transfer
        return d

    @classmethod
    def from_dict(cls, _input_dict):
        new_obj = cls()

        new_obj.identifier = _input_dict["identifier"]
        new_obj.display_name = _input_dict["display_name"]
        new_obj.percent_coverage = _input_dict["percent_coverage"]
        new_obj.in_conditioned_space = _input_dict["in_conditioned_space"]

        new_obj.energy_carrier = _input_dict["energy_carrier"]
        new_obj.solar_fraction = _input_dict["solar_fraction"]
        new_obj.util_fact_heat_transfer = _input_dict["util_fact_heat_transfer"]

        return new_obj


class PhHvacHotWaterHeaterHeatPump_Monthly(PhHvacHotWaterHeater):
    def __init__(self):
        super(PhHvacHotWaterHeaterHeatPump_Monthly, self).__init__()
        self.COP_1 = 2.5
        self.ambient_temp_1 = -8.333  # =17F
        self.COP_2 = 2.5
        self.ambient_temp_2 = 8.333  # =47F

    def to_dict(self, _include_properties=False):
        # type: (bool) -> Dict[str, Any]
        d = super(PhHvacHotWaterHeaterHeatPump_Monthly, self).to_dict()
        d["heater_type"] = self.__class__.__name__
        d["percent_coverage"] = self.percent_coverage
        d["in_conditioned_space"] = self.in_conditioned_space
        d["COP_1"] = self.COP_1
        d["ambient_temp_1"] = self.ambient_temp_1
        d["COP_2"] = self.COP_2
        d["ambient_temp_2"] = self.ambient_temp_2
        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict[str, Any]) -> PhHvacHotWaterHeaterHeatPump_Monthly
        new_obj = cls()
        new_obj.identifier = _input_dict["identifier"]
        new_obj.display_name = _input_dict["display_name"]
        new_obj.user_data = _input_dict.get("user_data", {})
        new_obj.percent_coverage = _input_dict["percent_coverage"]
        new_obj.in_conditioned_space = _input_dict["in_conditioned_space"]
        new_obj.COP_1 = _input_dict["COP_1"]
        new_obj.ambient_temp_1 = _input_dict["ambient_temp_1"]
        new_obj.COP_2 = _input_dict["COP_2"]
        new_obj.ambient_temp_2 = _input_dict["ambient_temp_2"]
        return new_obj


class PhHvacHotWaterHeaterHeatPump_Annual(PhHvacHotWaterHeater):
    def __init__(self):
        super(PhHvacHotWaterHeaterHeatPump_Annual, self).__init__()
        self.annual_COP = None  # type: (float | None)
        self.total_system_perf_ratio = None  # type: (float | None)

    def to_dict(self, _include_properties=False):
        # type: (bool) -> Dict[str, Any]
        d = super(PhHvacHotWaterHeaterHeatPump_Annual, self).to_dict()
        d["heater_type"] = self.__class__.__name__
        d["percent_coverage"] = self.percent_coverage
        d["in_conditioned_space"] = self.in_conditioned_space
        d["annual_COP"] = self.annual_COP
        d["total_system_perf_ratio"] = self.total_system_perf_ratio
        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict[str, Any]) -> PhHvacHotWaterHeaterHeatPump_Annual
        new_obj = cls()
        new_obj.identifier = _input_dict["identifier"]
        new_obj.display_name = _input_dict["display_name"]
        new_obj.user_data = _input_dict.get("user_data", {})
        new_obj.percent_coverage = _input_dict["percent_coverage"]
        new_obj.in_conditioned_space = _input_dict["in_conditioned_space"]
        new_obj.annual_COP = _input_dict["annual_COP"]
        new_obj.total_system_perf_ratio = _input_dict["total_system_perf_ratio"]
        return new_obj


class PhHvacHotWaterHeaterHeatPump_Inside(PhHvacHotWaterHeater):
    def __init__(self):
        super(PhHvacHotWaterHeaterHeatPump_Inside, self).__init__()
        self.annual_COP = None  # type: (float | None)
        self.total_system_perf_ratio = None  # type: (float | None)
        self.annual_energy_factor = None  # type: (float | None)

    def to_dict(self, _include_properties=False):
        # type: (bool) -> Dict[str, Any]
        d = super(PhHvacHotWaterHeaterHeatPump_Inside, self).to_dict()
        d["heater_type"] = self.__class__.__name__
        d["percent_coverage"] = self.percent_coverage
        d["in_conditioned_space"] = self.in_conditioned_space
        d["annual_COP"] = self.annual_COP
        d["total_system_perf_ratio"] = self.total_system_perf_ratio
        d["annual_energy_factor"] = self.annual_energy_factor
        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict[str, Any]) -> PhHvacHotWaterHeaterHeatPump_Inside
        new_obj = cls()
        new_obj.identifier = _input_dict["identifier"]
        new_obj.display_name = _input_dict["display_name"]
        new_obj.user_data = _input_dict.get("user_data", {})
        new_obj.percent_coverage = _input_dict["percent_coverage"]
        new_obj.in_conditioned_space = _input_dict["in_conditioned_space"]
        new_obj.annual_COP = _input_dict["annual_COP"]
        new_obj.total_system_perf_ratio = _input_dict["total_system_perf_ratio"]
        new_obj.annual_energy_factor = _input_dict["annual_energy_factor"]
        return new_obj


# -- Heater Builder -----------------------------------------------------------


class PhHvacHotWaterHeaterBuilder(object):
    """Constructor class for Hot-Water-Heater objects"""

    heaters = {
        "PhHvacHotWaterHeaterElectric": PhHvacHotWaterHeaterElectric,
        "PhHvacHotWaterHeaterBoiler": PhHvacHotWaterHeaterBoiler,
        "PhHvacHotWaterHeaterBoilerWood": PhHvacHotWaterHeaterBoilerWood,
        "PhHvacHotWaterHeaterDistrict": PhHvacHotWaterHeaterDistrict,
        "PhHvacHotWaterHeaterHeatPump_Annual": PhHvacHotWaterHeaterHeatPump_Annual,
        "PhHvacHotWaterHeaterHeatPump_Monthly": PhHvacHotWaterHeaterHeatPump_Monthly,
        "PhHvacHotWaterHeaterHeatPump_Inside": PhHvacHotWaterHeaterHeatPump_Inside,
    }

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (dict) -> PhHvacHotWaterHeater

        heater_type = _input_dict.get("heater_type")
        if heater_type is None:
            raise UnknownPhHvacHotWaterHeaterTypeError(cls.heaters.keys(), heater_type)

        heater_class = cls.heaters[heater_type]
        new_heater = heater_class.from_dict(_input_dict)

        return new_heater

    def __str__(self):
        return "{}()".format(self.__class__.__name__)

    def __repr__(self):
        return str(self)

    def ToString(self):
        return str(self)



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""Honeybee-Energy-PH Properties Extension for ServiceHotWater (Program) objects."""

try:
    from typing import Any, Optional
except:
    pass  # IronPython

try:
    from honeybee_energy.properties.extension import ServiceHotWaterProperties
except ImportError as e:
    raise ImportError("Failed to import honeybee_energy.", e)


class ServiceHotWaterPhProperties(object):
    """Honeybee-PH Properties for logging PH-style data."""

    def __init__(self, _host):
        # type: (Optional[ServiceHotWaterProperties]) -> None
        self._host = _host
        self.id_num = 0

    @property
    def host(self):
        # type: () -> Optional[ServiceHotWaterProperties]
        return self._host

    def to_dict(self, abridged=False, include=None):
        # type: (bool, Optional[list]) -> dict[str, Any]

        return {}

    @classmethod
    def from_dict(cls, _input_dict, _host):
        # type: (dict, Optional[ServiceHotWaterProperties]) -> ServiceHotWaterPhProperties
        new_prop = cls(_host)

        return new_prop

    def apply_properties_from_dict(self, abridged_data):
        # type: (Any) -> None
        return

    def __str__(self):
        return "{}()".format(self.__class__.__name__)

    def __repr__(self):
        """Properties representation."""
        return "{}".format(self.__class__.__name__)

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()
